function [R,prof] = profchol(R,prof,epsilon)
% ----------------------------------------------------------------------
% In-place profile (skyline) Cholesky decomposition of a real symmetric
% nxn matrix R. R is stored by columns and only its upper diagonal is
% accessed. prof is an n-element row vector specifying the profile shape
% of R. In column i, only elements from that specified by prof(i) down
% to the diagonal are accessed.  R and prof can be given and returned in
% one of three forms, prof being optional for the first two:
%
% 1. (The upper triangle of) a dense nxn matrix.  prof(i) is the index of
% the first nonzero entry in column i and only elements R(prof(i):i,i)
% are accessed.
%
% 2. A band matrix of above-diagonal bandwidth at most b<n. R is a bxn
% matrix representing the above-diagonal and diagonal entries of the
% unpacked input matrix, stored by columns. Only elements R(prof(i):b,i)
% are accessed, representing elements A(i+prof(i)-b:i,i) of the unpacked
% matrix. R(b,:) is the diagonal of A.
%
% 3. A fully packed profile matrix. R is a long column vector.  prof(i)
% gives the index of diagonal element i in R, the remainder of column i
% being stored immediately before this. I.e. R(prof(i-1)+1:prof(i))
% contains the nonzero entries A(i-prof(i)+prof(i-1)+1:i,i) of the
% unpacked input matrix. R(prof) is the diagonal of A.
%
% If epsilon>=0 is specified, diagonal elements of the output matrix
% that would be less than epsilon or imaginary are increased to
% epsilon. Otherwise imaginary entries are errors.
% ----------------------------------------------------------------------

[b,n] = size(R);
if nargin<2, prof=[]; end
if nargin<3, epsilon=-1; end
eps2 = max(0,epsilon)^2;

if b==n		% unpacked nxn matrix

   % Find profile if it wasn't given.
   if isempty(prof)
      prof = zeros(1,n);
      for i=1:n
	 nz = find([R(1:i,i);1]);
	 prof(i) = nz(1); 
      end
   end

   % Columnwise profile Cholesky.
   for i=1:n
      p = prof(i);
      for j=p:i-1
	 K = max(p,prof(j)):j-1;
	 R(j,i) = (R(j,i)-sum(R(K,j).*R(K,i))) / R(j,j);
      end
      a = R(i,i)-sum(R(p:i-1,i).^2);
      if a>=eps2,	 R(i,i) = sqrt(a); 
      elseif epsilon>=0, R(i,i) = epsilon; 
      else error(['profchol: matrix nonpositive at column ',int2str(i)]);
      end

      % dense version:
      % for i=1:n
      %   for j=1:i-1
      %     R(j,i) = (R(j,i)-sum(R(1:j-1,j).*R(1:j-1,i))) / R(j,j);
      %   end
      %   R(i,i) = sqrt(R(i,i)-sum(R(1:i-1,i).^2));
      % end
   end
   % R = triu(R);

elseif b<n		% packed band matrix

   % Find profile if it wasn't given.
   if isempty(prof)
      prof = zeros(1,n);
      for i=1:n 
	 nz = find([R(:,i);1]); 
	 prof(i) = nz(1);
      end
   end

   % Columnwise profile Cholesky. R(prof(k):b,k) represents elements
   % A(k+prof(k)-b:k,k) of unpacked matrix.
   for i=1:n
      p = prof(i);
      for j = p+i-b:i-1
	 K = max(p+i-j,prof(j)):b-1;
	 R(j-i+b,i) = (R(j-i+b,i)-sum(R(K,j).*R(K+(j-i),i))) / R(b,j);
      end
      a = R(b,i)-sum(R(p:b-1,i).^2);
      if a>=eps2,	 R(b,i) = sqrt(a); 
      elseif epsilon>=0, R(b,i) = epsilon; 
      else error(['profchol: matrix nonpositive at column ',int2str(i)]);
      end
   end

elseif n==1			% fully packed profile matrix

   if isempty(prof)
      error('profchol: no profile given for packed matrix');
   end
   n = size(prof,2);

   % Columnwise profile Cholesky. Column i is in R(prof(i-1)+1:prof(i)),
   % representing elements A(prof(i-1)+1-prof(i)+i,i) of unpacked
   % matrix. di = location of A(i,i) in R, ci = nominal offset of
   % element 0 of column i in R, pi = index of first nonzero element in
   % column i of A. Similarly for j.

   diprev = 0;
   for i=1:n
      di = prof(i);
      ci = di-i;
      pi = diprev+1-ci;
      if pi>1, djprev = prof(pi-1); else djprev = 0; end
      for j = pi:i-1
	 dj = prof(j);
	 cj = dj-j;
	 pj = djprev+1-cj;
	 K = cj+max(pi,pj):dj-1;
	 R(ci+j) = (R(ci+j)-sum(R(K).*R((ci-cj)+K))) / R(dj);
	 djprev = dj;
      end
      a = R(di)-sum(R(ci+pi:di-1).^2);
      if a>=eps2,	 R(di) = sqrt(a); 
      elseif epsilon>=0, R(di) = epsilon; 
      else error(['profchol: matrix nonpositive at column ',int2str(i)]);
      end
      diprev = di;
   end
   
else
   error('profchol: unkonwn matrix type');
end
